"""Implements the field of mines"""

import random
import time

import serge.actor
import serge.visual
import serge.sound
import serge.events
import serge.blocks.utils
import serge.blocks.animations
import serge.blocks.layout
import serge.blocks.visualblocks
import common
from common import G
import squares


class MineField(serge.actor.CompositeActor):
    """Implements the field of mines"""

    def __init__(self, tag, name, xs, ys, mines, triggers):
        """Initialise the minefield"""
        super(MineField, self).__init__(tag, name)
        #
        self.xs = xs
        self.ys = ys
        self.mines = mines
        self.triggers = triggers
        self.refresh_visual = False

    def getMinePositions(self):
        """Return all the possible mine positions and their probability"""
        positions = {}
        for x in range(self.xs):
            for y in range(self.ys):
                positions[(x, y)] = 1.0
        return positions

    def initMines(self):
        """Initialise the mines"""
        self.squares = []
        self.triggered_squares = serge.actor.ActorCollection()
        #
        # Calculate mine positions
        possible_mine_positions = self.getMinePositions()
        mine_positions = set()
        while len(mine_positions) < self.mines:
            chooser = serge.blocks.utils.ProbabilityChooser(possible_mine_positions)
            #
            # Pick a mine position
            x, y = chooser.choose()
            mine_positions.add((x, y))
            #
            # Make sure not to pick it again
            del(possible_mine_positions[(x, y)])
            #
            # If grouping then make the surrounding squares more likely to be picked
            if common.C_GAME_MODE == common.C_MODE_NEW:
                for dx in (-1, 0, +1):
                    for dy in (-1, 0, +1):
                        cx, cy = x + dx, y + dy
                        if (cx, cy) in possible_mine_positions:
                            possible_mine_positions[(cx, cy)] += G('mode-enhanced-grouping-factor')
        #
        # Calculate trigger positions
        trigger_positions = set()
        while len(trigger_positions) < self.triggers:
            x = random.randint(0, self.xs - 1)
            y = random.randint(0, self.ys - 1)
            if (x, y) not in mine_positions.union(trigger_positions):
                trigger_positions.add((x, y))
        #
        # Now put squares in
        for iy in range(self.ys):
            self.squares.append([])
            for ix in range(self.xs):
                new = squares.Mine if (ix, iy) in mine_positions \
                    else squares.Trigger if (ix, iy) in trigger_positions else squares.Square
                self.squares[-1].append(new(self.squares, (ix, iy)))
        #
        # The actual squares in the field
        for iy in range(self.ys):
            for ix in range(self.xs):
                actor = self.layout.addActor(
                    (ix, iy), SquareActor('square', 'square-%s-%s' % (ix, iy),
                                          self.squares[iy][ix], self),
                    'squares'
                )
                actor.linkEvent(common.E_BOMB_EXPLODED, self.bombExploded)
                actor.linkEvent(common.E_CELL_CLEARED, self.cellCleared)
                actor.linkEvent(common.E_FLAG_CLEARED, self.flagChanged, -1)
                actor.linkEvent(common.E_FLAG_PLACED, self.flagChanged, 1)
                actor.linkEvent(common.E_TRIGGER_HIT, self.triggerHit)
                actor.linkEvent(common.E_FIRST_TURN_EXPLOSION, self.firstTurnExplosion)
        #
        self.layout.updateRendering()
        #
        self.last_triggered = None
        self.trigger_timeout = G('trigger-timeout')

    def addedToWorld(self, world):
        """We were added to the world"""
        super(MineField, self).addedToWorld(world)
        #
        # The main layout
        self.layout = self.addChild(
            serge.blocks.layout.ManualUpdateGrid(
                'mine-grid', 'mine-grid', (self.xs, self.ys),
                width=self.xs * G('field-cell-size'),
                height=self.ys * G('field-cell-size')),
        )
        self.layout.setLayerName('main')
        self.layout.moveTo(*G('field-position'))
        #
        self.initMines()

    def bombExploded(self, obj, arg):
        """One of the bombs exploded"""
        serge.sound.Sounds.play('explode')
        for iy in range(self.ys):
            for ix in range(self.xs):
                self.layout.getActorAt((ix, iy)).revealBomb()
        self.processEvent((common.E_GAME_OVER, self))

    def cellCleared(self, square, arg):
        """A cell was cleared"""
        self.log.debug('Cell %s (%d, %d) was cleared' % (square, square.x, square.y))
        serge.sound.Sounds.play('click')
        common.GAME_TURNS += 1
        #
        # Reveal those around this square
        if square.getNeighbours() - square.getNumberNeighbourFlags() == 0:
            for dx in (-1, 0, 1):
                for dy in (-1, 0, 1):
                    cx, cy = square.x + dx, square.y + dy
                    if (cx < 0 or cy < 0 or cx >= self.xs or cy >= self.ys) or (dx == dy == 0):
                        pass
                    else:
                        cleared = self.layout.getActorAt((cx, cy))
                        if not cleared.revealed and not cleared.square.flagged:
                            cleared.squareReveal()
        #
        unknown = self.getNumberUnknown()
        if unknown <= self.mines:
            self.processEvent((common.E_FIELD_CLEARED, self))

    def flagChanged(self, obj, change):
        """A flag was changed"""
        serge.sound.Sounds.play('click')
        self.processEvent((common.E_FLAG_CHANGED, change))

    def resetField(self):
        """Reset the minefield to a new state"""
        self.layout.clearGrid()
        self.initMines()

    def getNumberUnknown(self):
        """Return the number of squares that are neither flagged nor revealed"""
        number = 0
        for row in self.squares:
            for cell in row:
                if not cell.revealed:
                    number += 1
        #
        return number

    def updateActor(self, interval, world):
        """Update the actor"""
        super(MineField, self).updateActor(interval, world)
        #
        # Watch out for needing to update the display
        if self.refresh_visual:
            self.layout.updateRendering()
            self.refresh_visual = False
        #
        # Watch for trigger expiring
        if self.triggered_squares and time.time() - self.last_triggered > self.trigger_timeout:
            self.triggerTimeout()
        #
        # Make sure cleared actors are not triggered - this can occur when clearing a
        # large area and recursively setting the trigger
        for actor in self.triggered_squares[:]:
            if actor.square.revealed or actor.square.flagged:
                actor.setTriggerState(False)
                self.triggered_squares.remove(actor)

    def triggerHit(self, square, arg):
        """A trigger was hit"""
        self.log.info('A trigger was hit on %s' % square.getNiceName())
        initial_triggers = len(self.triggered_squares)
        #
        # Add the triggered squares to the list we are tracking
        squares_triggered = square.square.getTriggeredSquares()
        for square in squares_triggered:
            actor = self.layout.getActorAt((square.x, square.y))
            if actor not in self.triggered_squares and actor.square.signal:
                self.triggered_squares.append(actor)
        #
        # Play a sound if we triggered something
        if initial_triggers == 0 and self.triggered_squares:
            serge.sound.Sounds.play('trigger')
        #
        # Set the triggers state for all the squares
        self.triggered_squares.forEach().setTriggerState(True)
        self.last_triggered = time.time()

    def triggerTimeout(self):
        """The trigger timed out"""
        self.log.info('The trigger has timed out')
        exploded = False
        for square in self.triggered_squares:
            if square.square.signal and not square.square.flagged and not square.square.revealed:
                square.explodeBomb()
                exploded = True
            square.setTriggerState(False)
        #
        self.triggered_squares = serge.actor.ActorCollection()
        if exploded:
            self.bombExploded(None, None)

    def firstTurnExplosion(self, square, arg):
        """Prevent a bomb going off on the first turn of the game

        It is a bit cruel to have the player fail on the first turn so if this would
        happen then we will move the bomb to another location. It is a cheat, but in
        the player's favour.

        """
        self.log.info('First turn explosion on square %s' % square.getNiceName())
        #
        # Find a place to move the bomb to
        while True:
            cell_x, cell_y = random.randint(0, self.xs - 1), random.randint(0, self.ys - 1)
            test_square = self.squares[cell_y][cell_x]
            if test_square.signal == 0 and test_square.trigger == 0:
                break
        #
        # Put a bomb there
        self.log.info('Switching bomb to be at %s, %s' % (cell_x, cell_y))
        bomb_square = squares.Mine(self.squares, (cell_x, cell_y))
        self.squares[cell_y][cell_x] = bomb_square
        self.layout.getActorAt((cell_x, cell_y)).setSquare(bomb_square)
        #
        # Set the offending square to not have a bomb on it
        cell_x, cell_y = square.square.x, square.square.y
        fresh_square = squares.Square(self.squares, (cell_x, cell_y))
        self.squares[cell_y][cell_x] = fresh_square
        square.setSquare(fresh_square)

    def getSquareAt(self, (x, y)):
        """Returns the square at a certain position"""
        try:
            cx, cy = self.layout.getCellFromLocation((x, y))
        except ValueError:
            return None
        else:
            return self.layout.getActorAt((cx, cy))


class SquareActor(serge.actor.MountableActor):
    """Represents a square on the screen"""

    def __init__(self, tag, name, square, parent):
        """Initialise the actor"""
        super(SquareActor, self).__init__(tag, name)
        #
        self.revealed = False
        self.parent = parent
        self.setSquare(square)
        #
        self.linkEvent(serge.events.E_LEFT_CLICK, self.squareClick)
        self.linkEvent(serge.events.E_RIGHT_CLICK, self.squareRightClick)

    def setSquare(self, square):
        """Set the square to use"""
        self.square = square
        self.normal_visual = square.sprite
        self.updateVisual(square.sprite)

    def addedToWorld(self, world):
        """Added to the world"""
        super(SquareActor, self).addedToWorld(world)
        #
        self.highlight = self.mountActor(
            serge.blocks.animations.AnimatedActor('highlight', 'highlight-%s' % self.name),
            (0, 0)
        )
        self.highlight.visual = serge.blocks.visualblocks.Rectangle(
            (G('field-cell-size'), G('field-cell-size')), G('field-cell-highlight-colour')
        )
        self.highlight.active = False
        self.highlight.visible = True
        self.highlight.setLayerName('highlight')
        self.highlight.addRegisteredAnimation('pulsed-visibility')

    def updateVisual(self, sprite_name):
        """Update our visual"""
        if self.revealed:
            if common.C_GAME_MODE == common.C_MODE_CLASSIC or G('mode-enhanced-show-numbers'):
                overlay_text = str(self.neighbours) if self.neighbours != 0 else ' '
            else:
                overlay_text = '+' if self.neighbours != 0 else ' '
        else:
            overlay_text = ''
        self.visual = serge.blocks.visualblocks.SpriteText(
            overlay_text,
            G('field-cell-text-colour'),
            sprite_name
        )
        self.parent.refresh_visual = True

    def revealNumber(self):
        """Reveal the number of mines close to us"""
        self.revealed = True
        self.neighbours = self.square.getNeighbours()
        self.normal_visual = 'cleared'
        self.updateVisual('cleared')
        self.processEvent((common.E_CELL_CLEARED, self.square))

    def revealBomb(self):
        """Reveal a bomb if we have one"""
        if self.square.signal and not self.revealed and not self.square.exploded:
            self.updateVisual('bomb')

    def explodeBomb(self):
        """Explode a bomb"""
        if self.square.signal:
            self.updateVisual('exploded')
            self.square.exploded = True

    def squareClick(self, obj, arg):
        """Clicked on this square"""
        self.log.debug('Clicked on square %s' % self.getNiceName())
        if common.GAME_OVER:
            return
        #
        # Watch for exploding a bomb on the first turn - this is no
        # fun (!) so we will do a little cheat to avoid the situation
        if common.GAME_TURNS == 0 and self.square.signal:
            self.processEvent((common.E_FIRST_TURN_EXPLOSION, self))
        #
        self.squareReveal()

    def squareReveal(self):
        """Reveal this square"""
        #
        # Did we hit a bomb?
        if self.square.signal:
            self.revealed = self.square.revealed = True
            self.neighbours = 0
            self.explodeBomb()
            self.processEvent((common.E_BOMB_EXPLODED, self))
        else:
            if self.square.flagged:
                self.squareRightClick(None, None)
            #
            # If already revealed then check if the number of flags is the number
            # of mines - if so then we effectively clear the square
            if self.revealed:
                if self.square.getNumberNeighbourFlags() == self.square.getNeighbours():
                    self.log.info('Clearing a square by flag count')
                    self.processEvent((common.E_CELL_CLEARED, self.square))
            else:
                self.revealed = self.square.revealed = True
                self.revealNumber()
                if self.square.trigger and self.neighbours != 0:
                    self.log.debug('Stepped on a trigger')
                    self.processEvent((common.E_TRIGGER_HIT, self))

    def squareRightClick(self, obj, arg):
        """Right clicked on this square"""
        if common.GAME_OVER or self.revealed:
            return
        if self.square.flagged:
            self.updateVisual(self.normal_visual)
            self.square.flagged = False
            self.processEvent((common.E_FLAG_CLEARED, self))
        elif common.HAVE_FLAGS:
            self.updateVisual('flag')
            self.square.flagged = True
            self.processEvent((common.E_FLAG_PLACED, self))
        else:
            serge.sound.Sounds.play('no-flag')

    def setTriggerState(self, state):
        """Set as triggered"""
        self.highlight.active = self.highlight.visible = state