"""Main module for the game"""

import pygame
import os
import time

import serge.engine
import serge.world
import serge.actor
import serge.zone
import serge.render
import serge.sound 
import serge.events
import serge.blocks.achievements
try:
    import serge.builder.builder
except ImportError:
    pass

import serge.visual
import serge.blocks.actors
import serge.blocks.utils
import serge.blocks.animations
import serge.blocks.onlinescores

import common
import mainscreen
import startscreen
import helpscreen
import creditsscreen
import storage
import highscores

from theme import G, theme


def registerSounds():
    """Register the sounds to use"""
    serge.sound.Sounds.setPath('sound')
    r = serge.sound.Sounds.registerItem
    r('click', '164643__adam-n__pen-click-3.wav')
    r('button', '146775__nlm__cooker-hood-fan-light-switch.wav')
    r('screen', '125326__adeluc4__camera-flash-charge.wav')
    r('select', '163401__baryy__click.wav')
    r('power', '42943__agfx__lrg-push-butten-switch-2-l.wav')
    r('explode', '110391__soundscalpel-com__firework-explosion-fizz.wav')
    r('no-flag', '20794__djgriffin__empty-beer-can.wav')
    r('trigger', '104408__nanashi__slide-released.wav')


def registerMusic():
    """Register the music to use"""
    serge.sound.Music.setPath('music')
    r = serge.sound.Music.registerItem


def registerGraphics(cheat):
    """Register the graphics to use"""
    serge.visual.Sprites.setPath('graphics')
    r = serge.visual.Sprites.registerItem
    rf = serge.visual.Sprites.registerFromFiles
    rm = serge.visual.Sprites.registerMultipleItems
    rp = serge.visual.Sprites.registerItemsFromPattern
    r('icon', 'icon.png')
    r('logo', 'logo.png', zoom=0.5)
    r('small-icon', 'icon.png', zoom=0.2)
    r('help-text', 'help-text.png')
    rf('mute-button', 'music-%d.png', 2, zoom=0.5)
    rf('achievement', 'achievement-%d.png', 2)
    r('square', 'square.png', convert_alpha=True)
    r('cleared', 'cleared.png', convert_alpha=True)
    r('bomb-square', 'bomb-square.png' if cheat else 'square.png', convert_alpha=True)
    r('trigger-square', 'trigger-square.png' if cheat else 'square.png', convert_alpha=True)
    r('bomb', 'bomb.png', convert_alpha=True)
    r('exploded', 'exploded.png', convert_alpha=True)
    r('flag', 'flag.png', convert_alpha=True)
    r('restart', 'restart.png', convert_alpha=True)
    r('overlay', 'overlay.png', convert_alpha=True)
    r('red-button', 'red-button.png')
    r('yellow-button', 'yellow-button.png')
    r('green-right-on-button', 'green-right-on.png')
    r('green-right-off-button', 'green-right-off.png')
    r('green-left-on-button', 'green-left-on.png')
    r('green-left-off-button', 'green-left-off.png')
    r('on-off-button', 'on-off-button.png')
    r('scope', 'scope.png')
    rp('clock\-(\w+)\.png', convert_alpha=True)
    #
    # Fonts
    serge.visual.Fonts.setPath('fonts')
    serge.visual.Fonts.registerItem('led', 'RADIOLAND.ttf')
    serge.visual.Fonts.registerItem('terminal', 'NonFiction.ttf')


def registerEvents():
    """Register all the events"""
    broadcaster = serge.events.getEventBroadcaster()
    broadcaster.registerEventsFromModule(common)


def registerAchievements(options):
    """Register the achievements for the game"""
    r = serge.blocks.achievements.initManager('minefield').safeRegisterAchievement
    a = serge.blocks.achievements.Achievement
    r(a('Played the game', 'You played the game at least once', 
        'badge', False, 'play', condition_string=': True'))
    serge.blocks.achievements.addAchievementsWorld(options, theme)
    

def startEngine(options):
    """Start the main engine"""
    engine = serge.engine.Engine(
        width=G('screen-width'), height=G('screen-height'),
        title=G('screen-title'), icon=os.path.join('graphics', G('screen-icon-filename')))
    serge.blocks.utils.createVirtualLayersForEngine(
        engine, ['background', 'foreground', 'main', 'squares', 'highlight', 'ui',
                 'confirm', 'overlay', 'buttons', 'debug'])
    serge.blocks.utils.createWorldsForEngine(
        engine,
        ['start-screen', 'main-screen', 'credits-screen', 'help-screen', 'high-score-screen'])
    #
    engine.setCurrentWorldByName('main-screen')
    return engine


def registerAnimations():
    """Register the animations to use"""
    r = serge.blocks.animations.Animations.registerItem
    r(
        'pulsed-zoom',
        serge.blocks.animations.PulseZoom(
            G('button-zoom-start'), G('button-zoom-end'), G('button-zoom-duration'),
            loop=True),
    )
    r(
        'pulsed-visibility',
        serge.blocks.animations.PulsedVisibility(
            G('high-score-duration'),
            G('high-score-fraction')
        ),
    )


def stoppingNow(obj, arg):
    """We are about to stop"""
    #
    # Fade out music and wait for a bit before going away
    serge.sound.Music.fadeout(G('pre-stop-pause')*1000)
    time.sleep(G('pre-stop-pause'))
       

def createHighScores(options):
    """Create the high score table"""
    hs = serge.blocks.onlinescores.HighScoreSystem(
        G('app-url', 'high-score-screen'), secret_user=options.cheat)
    app_name = G('app-name', 'high-score-screen')
    if hs.gameExists(app_name):
        return
    #
    common.log.info('Creating high score table')
    hs.createGame(app_name)
    time.sleep(1.0)
    for difficulty in G('difficulties').keys():
        for mode, _, _, _ in G('game-mode-buttons'):
            hs.createGameCategory(
                app_name,
                '%s - %s' % (difficulty, mode),
                'time',
                True,
                10
            )


def main(options, args):
    """Start the engine and the game"""
    #
    # Create the high scores
    if options.high_score:
        createHighScores(options)
    #
    # Create the engine
    engine = startEngine(options)
    engine.linkEvent(serge.events.E_BEFORE_STOP, stoppingNow)
    #
    registerAnimations()
    registerSounds()
    registerMusic()
    registerGraphics(options.cheat)
    registerEvents()
    #
    storage.initStorage()
    last_user = storage.User.getLastUser()
    #
    # Record a movie
    if options.movie:
        serge.blocks.utils.RecordDesktop(options.movie)
    #
    # Change theme settings
    if options.theme:
        theme.updateFromString(options.theme)
    #
    # Muting
    mute = serge.blocks.actors.MuteButton('mute-button', 'buttons', alpha=G('mute-button-alpha'))
    serge.blocks.utils.addMuteButtonToWorlds(mute, center_position=G('mute-button-position'))
    #
    if options.musicoff:
        mute.toggleSound()
    #
    # Initialise the main logic
    registerAchievements(options)
    mainscreen.main(options)
    startscreen.main(options)
    helpscreen.main(options)
    creditsscreen.main(options)
    highscores.main(options)
    #   
    if options.debug:
        serge.builder.builder.main(engine, options.framerate)
    else:
        engine.run(options.framerate)
