"""Implements a simple oscilloscope to give an indication where the mines are"""

import pygame
import math
import random

import serge.actor
import serge.visual


class Oscilloscope(serge.actor.CompositeActor):
    """A simple oscilloscope"""

    def __init__(self, tag, name, size, colour,
                 x_points, x_width, relaxation, angle_jitter, amplitude_jitter, y_offset):
        """Initialise the oscilloscope"""
        super(Oscilloscope, self).__init__(tag, name)
        #
        self.colour = colour
        self.x_points = x_points
        self.x_width = x_width
        self.relaxation = relaxation
        self.angle_jitter = angle_jitter
        self.amplitude_jitter = amplitude_jitter
        self.y_offset = y_offset
        self.visual = serge.visual.SurfaceDrawing(size[0], size[1])
        self.angles = []
        self.values = [0.0 for _ in range(x_points)]

    def setAngles(self, angles):
        """Set the angles"""
        self.angles = angles

    def updateActor(self, interval, world):
        """Update the actor"""
        super(Oscilloscope, self).updateActor(interval, world)
        #
        # Clear the old data
        self.visual.clearSurface()
        surface = self.visual.getSurface()
        point_width = self.width / self.x_points
        relaxation = min(1.0, self.relaxation * interval / 1000.0)
        angle_jitter_offset = random.randrange(-self.angle_jitter, self.angle_jitter)
        #
        # Re-plot the new
        for x in range(self.x_points):
            amplitude_jitter_offset = (random.random() - 0.5) * self.amplitude_jitter * 2.0
            #
            # Get the target new value
            px = float(x) / (self.x_points - 1) * self.width
            vx = float(x) / (self.x_points - 1) * 360.0 - 180.0 + angle_jitter_offset
            py = self.height - self.getValue(vx) * self.height + amplitude_jitter_offset + self.y_offset
            #
            # Under-relax the value
            self.values[x] = self.values[x] * (1.0 - relaxation) + py * relaxation
            pygame.draw.rect(surface, self.colour, (px, self.values[x], point_width + 1, point_width + 1))

    def getValue(self, x):
        """Return the value to use"""
        if self.angles:
            parts = [
                math.cos(math.radians(x - angle)) / (1.0 + ((x - angle) / self.x_width) ** 2) +
                math.cos(math.radians(x + 360 - angle)) / (1.0 + ((x + 360 - angle) / self.x_width) ** 2)
                for angle in self.angles]
            return sum(parts) / len(parts)
        else:
            return 0.0